from .base_template import BaseTemplate
from .header import HeaderStyle, HeaderTemplate
from .toc import ToCStyle, ToCTemplate

__all__ = [
    "BaseTemplate",
    "HeaderTemplate",
    "HeaderStyle",
    "ToCTemplate",
    "ToCStyle",
]



"""
Pydantic models and settings for the readme-ai package.
"""

from __future__ import annotations

from enum import Enum
from functools import cached_property
from pathlib import Path
from typing import Literal

from pydantic import (
    AnyHttpUrl,
    BaseModel,
    ConfigDict,
    Field,
    FilePath,
    PositiveFloat,
    PositiveInt,
    field_validator,
    model_validator,
)
from pydantic_extra_types.color import Color

from readmeai.core.logger import Logger
from readmeai.utils.file_handler import FileHandler
from readmeai.utils.file_resources import get_resource_path
from readmeai.vcs.errors import GitValidationError
from readmeai.vcs.url_builder import GitURL, parse_git_url

_logger = Logger(__name__)


class BadgeOptions(str, Enum):
    DEFAULT = "default"
    FLAT = "flat"
    FLAT_SQUARE = "flat-square"
    FOR_THE_BADGE = "for-the-badge"
    PLASTIC = "plastic"
    SKILLS = "skills"
    SKILLS_LIGHT = "skills-light"
    SOCIAL = "social"


class ImageOptions(str, Enum):
    CUSTOM = "CUSTOM"
    LLM = "LLM"
    BLACK = "https://img.icons8.com/external-tal-revivo-regular-tal-revivo/96/external-readme-is-a-easy-to-build-a-developer-hub-that-adapts-to-the-user-logo-regular-tal-revivo.png"
    BLUE = "https://raw.githubusercontent.com/PKief/vscode-material-icon-theme/ec559a9f6bfd399b82bb44393651661b08aaf7ba/icons/folder-markdown-open.svg"
    CLOUD = "https://cdn-icons-png.flaticon.com/512/6295/6295417.png"
    GRADIENT = "https://img.icons8.com/?size=512&id=55494&format=png"
    GREY = "https://img.icons8.com/external-tal-revivo-filled-tal-revivo/96/external-markdown-a-lightweight-markup-language-with-plain-text-formatting-syntax-logo-filled-tal-revivo.png"
    PURPLE = "https://img.icons8.com/external-tal-revivo-duo-tal-revivo/100/external-markdown-a-lightweight-markup-language-with-plain-text-formatting-syntax-logo-duo-tal-revivo.png"


class ModelOptions(str, Enum):
    # ANTHROPIC = "ANTHROPIC"
    GEMINI = "GEMINI"
    OFFLINE = "OFFLINE"
    OLLAMA = "OLLAMA"
    OPENAI = "OPENAI"


class APISettings(BaseModel):
    """
    LLM API settings and parameters.
    """

    rate_limit: PositiveInt = Field(gt=0, le=25, description="API rate limit.")
    system_message: str = Field(description="LLM system prompt content field.")


class FileSettings(BaseModel):
    """
    File path resources for the readme-ai package.
    """

    commands: str = Field(description="'Quickstart' setup commands.")
    ignore_list: str = Field(description="List of files to ignore.")
    languages: str = Field(description="Extension to language mappings.")
    markdown: str = Field(description="Markdown code template blocks.")
    parsers: str = Field(description="Common dependency file names.")
    prompts: str = Field(description="LLM API prompt templates.")
    shieldsio_icons: str = Field(description="Shields.io svg icon badges.")
    skill_icons: str = Field(description="Skill icon badges.")


class GitSettings(BaseModel):
    """
    User repository settings for a remote or local codebase.
    """

    repository: Path | str
    git_url: GitURL | None = None
    full_name: str | None = None
    host_domain: str | None = None
    host: str | None = None
    name: str | None = None

    model_config = ConfigDict(extra="forbid")

    @field_validator("repository")
    def validate_repository(cls, value: Path | str) -> Path | str:
        """
        Validates the repository path or Git URL.
        """
        if isinstance(value, Path) or (
            isinstance(value, str)
            and Path(value).is_dir()
            and Path(value).exists()
        ):
            return value
        try:
            return str(GitURL.create(value).url)
        except ValueError as exc:
            raise GitValidationError(
                f"Invalid Git repository URL or path: {value}",
            ) from exc

    @field_validator("git_url")
    def set_git_url(cls, value: GitURL | None, values) -> GitURL | None:
        """
        Set Git URL from the repository path if not provided.
        """
        if value is None and "repository" in values:
            return GitURL.create(str(values["repository"]))
        return value

    @model_validator(mode="after")
    def set_git_attributes(self) -> GitSettings:
        """
        Parse and set Git repository attributes.
        """
        if self.git_url:
            self.host_domain = self.git_url.host_domain
            self.host = self.git_url.host.name if self.git_url.host else None
            self.name = self.git_url.name
            self.full_name = self.git_url.full_name
        else:
            self.host_domain, self.host, self.name, self.full_name = (
                parse_git_url(str(self.repository))
            )
        return self

    def get_file_url(self, file_path: str) -> str:
        """
        Generates a URL for a file in the repository.
        """
        if self.git_url:
            return self.git_url.get_file_url(file_path)
        elif isinstance(self.repository, Path) or (
            isinstance(self.repository, str) and Path(self.repository).is_dir()
        ):
            return str(Path(self.repository) / file_path)
        else:
            raise GitValidationError("Failed to generate git file URL")


class MarkdownSettings(BaseModel):
    """
    Markdown code template blocks for building the README.md file.
    """

    align: Literal["left", "center", "right"] = Field(
        default="center",
        description="align for markdown content.",
    )
    badge_color: Color = Field(
        default_factory=lambda: Color("blue"),
        description="Badge color (https://www.w3.org/TR/SVG11/types.html#ColorKeywords)",
    )
    badge_style: BadgeOptions = Field(
        default=BadgeOptions.FLAT,
        description="Badge icon style type.",
    )
    badge_icons: str
    contribute: str
    emojis: bool = Field(
        default=False,
        description="Enable emoji prefixes for headers.",
    )
    features: str = Field(
        default="❯ INSERT-PROJECT-FEATURES",
        description="Project feature content.",
    )
    header_style: str = Field(
        default="classic",
        description="Header style for the README file.",
    )
    image: AnyHttpUrl | FilePath | str = Field(
        default=ImageOptions.BLUE,
        description="Image URL or path for the project logo.",
    )
    image_width: str = Field(
        default="20%",
        description="Project logo width",
    )
    modules: str
    modules_widget: str
    overview: str = Field(
        default="❯ INSERT-PROJECT-OVERVIEW",
        description="Project overview content.",
    )
    placeholder: str = Field(
        default="<code>❯ REPLACE-ME</code>",
        description="Placeholder image for missing content.",
    )
    quickstart: str
    requirements: str = Field(
        default="",
        description="Project system prerequisites.",
    )
    shieldsio_icons: str
    skill_icons: str
    slogan: str = Field(
        default="❯ INSERT-PROJECT-SLOGAN",
        description="Project tagline or slogan.",
    )
    tables: str = Field(default="", description="Markdown table options.")
    toc_style: str = Field(
        default="bullet",
        description="Table of contents content.",
    )
    tree: str
    tree_depth: PositiveInt = Field(
        default=2,
        ge=1,
        le=5,
        description="Depth of directory tree.",
    )

    @field_validator("badge_color")
    def set_color(cls, value: str) -> str:
        """
        Validates badge color value and returns the hex code.
        """
        try:
            return Color(value).as_hex().strip("#")
        except ValueError as exc:
            _logger.error(f"Invalid color value '{value}': {exc}")
            return cls.model_fields["badge_color"].default

    @model_validator(mode="after")
    def set_width(self) -> MarkdownSettings:
        """
        Validates and sets the width for the project logo image.
        """
        if str(self.image).lower() == ImageOptions.LLM.name.lower():
            self.image_width = "60%"
        return self


class ModelSettings(BaseModel):
    """
    LLM API model settings and parameters.
    """

    api: str | None = Field(
        default=ModelOptions.OFFLINE,
        description="API key for the LLM model.",
    )
    base_url: str
    context_window: PositiveInt
    encoder: str
    host_name: AnyHttpUrl
    localhost: AnyHttpUrl
    model: str
    path: str
    temperature: float
    tokens: PositiveInt
    top_p: PositiveFloat


class Settings(BaseModel):
    """
    Pydantic settings model for the readme-ai package.
    """

    api: APISettings
    files: FileSettings
    git: GitSettings
    llm: ModelSettings
    md: MarkdownSettings

    model_config = ConfigDict(
        validate_assignment=True,
    )


class ConfigLoader:
    """
    Loads the configuration settings for the readme-ai package.
    """

    def __init__(
        self,
        config_file: str = "config.toml",
        submodule: str = "settings",
    ) -> None:
        """Initialize ConfigLoader with the base configuration file."""
        self.file_handler = FileHandler()
        self.config_file = config_file
        self.submodule = submodule
        self.config = self._load_config
        self.load_settings()

    @cached_property
    def _load_config(self) -> Settings:
        """Loads the base configuration file."""
        file_path = get_resource_path(
            file_path=self.config_file,
            submodule=self.submodule,
        )
        config_dict = self.file_handler.read(str(file_path))
        return Settings.model_validate(config_dict)

    def load_settings(self) -> dict[str, dict]:
        """
        Loads all configuration settings.
        1. Loads the base configuration file from 'settings/config.toml'.
        2. Loads all additional TOML files defined in 'FileSettings.'
        """
        settings = self._load_config.model_dump()

        for key, file_path in settings["files"].items():
            if not file_path.endswith(".toml"):
                continue

            file_path = get_resource_path(file_path=file_path)
            data_config = self.file_handler.read(file_path)
            settings[key] = data_config
            setattr(self, key, data_config)

            _logger.info(f"Config loaded: {self.submodule}/{file_path}")

        return settings



"""
Custom exceptions for the utilities package.
"""

from __future__ import annotations

from readmeai._exceptions import ReadmeAIError


class GitValidationError(ReadmeAIError):
    """
    Base class errors validating Git repositories.
    """

    ...


class GitCloneError(GitValidationError):
    """
    Raised when a Git repository cannot be cloned.
    """

    def __init__(self, repository: str, *args):
        self.repository = repository
        super().__init__(f"Failed to clone repository: {repository}", *args)


class GitURLError(GitValidationError):
    """
    Raised when an invalid Git repository URL is provided.
    """

    def __init__(self, url: str, *args):
        self.url = url
        super().__init__(f"Invalid Git repository URL: {url}", *args)


class UnsupportedGitHostError(GitValidationError):
    """
    Raised when an unsupported Git host is provided.
    """

    def __init__(self, host: str, *args):
        self.host = host
        super().__init__(f"Unsupported Git host: {host}", *args)



from .errors import (
    GitCloneError,
    GitURLError,
    GitValidationError,
    UnsupportedGitHostError,
)
from .ingestor import retrieve_repository
from .metadata import (
    RepositoryMetadata,
    _fetch_repository_metadata,
    fetch_git_repository_metadata,
)
from .providers import GitHost, get_file_url, parse_git_url
from .url_builder import GitURL

__all__ = [
    "GitHost",
    "GitURL",
    "GitURL.validate_url",
    "GitURL.set_attributes",
    "GitURL.create",
    "GitURL.git_file_url",
    "GitURL.git_api_url",
    "get_file_url",
    "parse_git_url",
    "retrieve_repository",
    "GitValidationError",
    "GitCloneError",
    "GitURLError",
    "UnsupportedGitHostError",
    "RepositoryMetadata",
    "fetch_git_repository_metadata",
    "_fetch_repository_metadata",
]



"""
Builds each section of the README Markdown file.
"""

__package__ = "readmeai"

from pathlib import Path

from readmeai.config.settings import BadgeOptions, ConfigLoader
from readmeai.generators import badges, tables, tree, utils
from readmeai.generators.quickstart import get_setup_data
from readmeai.templates.header import HeaderTemplate
from readmeai.templates.toc import ToCTemplate
from readmeai.vcs.providers import GitHost


class MarkdownBuilder:
    """Builds each section of the README Markdown file."""

    def __init__(
        self,
        config_loader: ConfigLoader,
        dependencies: list[str],
        summaries: list[tuple[str, str]],
        temp_dir: str,
    ):
        self.deps = dependencies
        self.summaries = summaries
        self.temp_dir = Path(temp_dir)
        self.config = config_loader.config
        self.config_loader = config_loader
        self.md = self.config.md
        self.git = self.config.git
        self.repo_url = (
            self.git.repository
            if self.git.host_domain != GitHost.LOCAL.name.lower()
            else f"../{self.git.name}"
        )
        self.header_template = HeaderTemplate(self.md.header_style)
        self.toc_template = ToCTemplate(self.md.toc_style)

    @property
    def md_header(self) -> str:
        """
        Generates the README header section.
        """
        if BadgeOptions.SKILLS.value not in self.md.badge_style:
            md_shields, md_badges = badges.shieldsio_icons(
                self.config,
                self.deps,
                str(self.git.full_name),
                str(self.git.host),
            )
        else:
            md_shields = (
                "<!-- Shields.io badges disabled, using skill icons. -->"
            )
            md_badges = badges.skill_icons(self.config, self.deps)

        header_data = {
            "align": self.md.align,
            "image": self.md.image,
            "image_width": self.md.image_width,
            "repo_name": self.git.name.upper()
            if self.git.name
            else self.md.placeholder,
            "slogan": self.md.slogan,
            "shields_icons": md_shields,
            "badge_icons": md_badges,
        }
        return self.header_template.render(header_data)

    @property
    def md_toc(self) -> str:
        """
        Generates the README Table of Contents section.
        """
        toc_data = {
            "sections": [
                {"title": "📍 Overview"},
                {"title": "👾 Features"},
                {"title": "📂 Repository Structure"},
                {"title": "🧩 Modules"},
                {
                    "title": "🚀 Getting Started",
                    "subsections": [
                        {"title": "🔖 Prerequisites"},
                        {"title": "📦 Installation"},
                        {"title": "🤖 Usage"},
                        {"title": "🧪 Tests"},
                    ],
                },
                {"title": "📌 Project Roadmap"},
                {"title": "🤝 Contributing"},
                {"title": "🎗 License"},
                {"title": "🙌 Acknowledgments"},
            ],
        }
        return self.toc_template.render(toc_data)

    @property
    def md_summaries(self) -> str:
        """Generates the README code summaries section."""
        summaries = tables.format_code_summaries(
            self.md.placeholder,
            self.summaries,
        )
        md_summaries = tables.generate_markdown_tables(
            self.md.modules_widget,
            summaries,
            self.git.full_name,
            self.repo_url,
        )
        return md_summaries

    @property
    def md_tree(self) -> str:
        """Generates the README directory tree structure."""
        md_tree = tree.TreeGenerator(
            repo_name=self.git.name,
            root_dir=self.temp_dir,
            repo_url=self.repo_url,
            max_depth=self.md.tree_depth,
        ).tree()
        return self.md.tree.format(md_tree)

    @property
    def md_quickstart(self) -> str:
        """Generates the README Getting Started section."""
        setup_data = get_setup_data(self.config_loader, self.summaries)
        return self.md.quickstart.format(
            repo_name=self.git.name,
            repo_url=self.repo_url,
            prerequisites=setup_data.prerequisites,
            install_command=setup_data.install_command,
            run_command=setup_data.run_command,
            test_command=setup_data.test_command,
        )

    @property
    def md_contributing(self) -> str:
        """Generates the README Contributing section."""
        return self.md.contribute.format(
            host=self.git.host,
            host_domain=self.git.host_domain,
            full_name=self.git.full_name,
            repo_name=self.git.name,
            repo_url=self.repo_url,
        )

    def build(self) -> str:
        """
        Builds each section of the README.md file.
        """
        md_contents = [
            self.md_header,
            self.md_toc,
            self.md.overview,
            self.md.features,
            self.md_tree,
            self.md.modules,
            self.md_summaries,
            self.md_quickstart,
            self.md_contributing,
        ]

        if self.config.md.emojis is False:
            md_contents = utils.remove_emojis(md_contents)

        return "\n".join(md_contents)



import re
from enum import Enum
from typing import Any, ClassVar

from .base_template import BaseTemplate


class ToCStyle(Enum):
    """
    Enum for Table of Contents template styles.
    """

    BULLET = "bullet"
    FOLD = "fold"
    LINKS = "links"
    NUMBER = "number"


class ToCTemplate(BaseTemplate):
    """
    Class variable for rendering the README.md Table of Contents.
    """

    TOC_TEMPLATES: ClassVar[dict] = {
        ToCStyle.BULLET: """##### 🔗 Table of Contents\n\n{toc_items}\n---\n""",
        ToCStyle.FOLD: """<details><summary>Table of Contents</summary>\n\n{toc_items}\n</details>\n<hr>\n""",
        ToCStyle.NUMBER: """##### 📖 Table of Contents\n\n{toc_items}\n---\n""",
        ToCStyle.LINKS: """##### 🔗 Quick Links\n\n{toc_items}\n---\n""",
    }

    TOC_ITEM_TEMPLATES: ClassVar[dict[ToCStyle, str]] = {
        ToCStyle.BULLET: "- [{title}](#{anchor})\n",
        ToCStyle.FOLD: "- [{title}](#{anchor})\n",
        ToCStyle.NUMBER: "{number}. [{title}](#{anchor})\n",
        ToCStyle.LINKS: "- [{title}](#{anchor})\n",
    }

    def __init__(self, style: str = ToCStyle.BULLET.value) -> None:
        """
        Initialize the Table of Contents template with the given style.
        """
        self.style = ToCStyle(style)

    @staticmethod
    def generate_anchor(title: str) -> str:
        """
        Generate an anchor link from the given title.
        """
        title = re.sub(r":[a-zA-Z0-9_+-]+:", "", title)
        anchor = title.lower().replace(" ", "-")
        anchor = re.sub(r"[^\w-]", "", anchor)
        return anchor

    def _generate_toc_items(
        self,
        sections: list[dict[str, Any]],
        level: int = 0,
        parent_number: str = "",
    ) -> str:
        """
        Generate Table of Contents items recursively.
        """
        toc = ""
        for index, section in enumerate(sections, start=1):
            indent = "  " * level
            anchor = self.generate_anchor(section["title"])
            if self.style == ToCStyle.NUMBER:
                number = (
                    f"{parent_number}{index}" if parent_number else str(index)
                )
                item = self.TOC_ITEM_TEMPLATES[self.style].format(
                    number=number,
                    title=section["title"],
                    anchor=anchor,
                )
            else:
                item = self.TOC_ITEM_TEMPLATES[self.style].format(
                    title=section["title"],
                    anchor=anchor,
                )
            toc += f"{indent}{item}"
            if "subsections" in section:
                toc += self._generate_toc_items(
                    section["subsections"],
                    level + 2,
                    f"{number}." if self.style == ToCStyle.NUMBER else "",
                )
        return toc

    def render(self, data: dict[str, Any]) -> str:
        """
        Render Table of Contents based on the current style and data.
        """
        toc_items = self._generate_toc_items(data["sections"])
        template = self.TOC_TEMPLATES[self.style]
        return template.format(toc_items=toc_items)

    @staticmethod
    def get_toc_template(template: str) -> str:
        """
        Get the Table of Contents template for the given style.
        """
        return ToCTemplate.TOC_TEMPLATES[ToCStyle(template)]



"""
Dynamically generate 'Quickstart' guides for the README file.
"""

import traceback
from dataclasses import dataclass
from pathlib import Path

from readmeai.config.settings import ConfigLoader
from readmeai.core.logger import Logger

_logger = Logger(__name__)


@dataclass
class QuickStart:
    """
    Information about using, running, and testing a repository.
    """

    install_command: str
    run_command: str
    test_command: str
    prerequisites: str
    language_counts: dict[str, int]
    language_key: str | None
    language_name: str | None = None


def count_languages(
    summaries: tuple,
    config_loader: ConfigLoader,
) -> dict[str, int]:
    """
    Counts the occurrences of each language in the summaries.
    """
    parser_files = config_loader.parsers.get("parsers")

    language_counts: dict[str, int] = {}

    for file_path, _ in summaries:
        language = Path(file_path).suffix[1:] or None

        if str(file_path) in [
            dependency_file for dependency_file in parser_files
        ]:
            continue

        if (
            language
            and language.strip()
            and language not in config_loader.ignore_list
        ):
            language_counts[language] = language_counts.get(language, 0) + 1

    return language_counts


def get_top_language(language_counts: dict[str, int]) -> str | None:
    """
    Determines the top language.
    """
    if not language_counts:
        return None
    else:
        return max(sorted(language_counts), key=language_counts.get)


def get_top_language_setup(
    language_counts: dict,
    config_loader: ConfigLoader,
) -> QuickStart:
    """
    Determines the top language and retrieves its setup commands.
    """
    languages = config_loader.languages.get("language_names")
    commands = config_loader.commands.get("quickstart_guide")

    language_key = get_top_language(language_counts)
    language_name = languages.get(language_key, languages.get("default"))
    quickstart_commands = commands.get(language_name, commands.get("default"))
    prerequisites = f"**{language_name}**: `version x.y.z`"

    return QuickStart(
        *quickstart_commands,
        prerequisites=prerequisites,
        language_counts=language_counts,
        language_key=language_key,
        language_name=language_name,
    )


def get_setup_data(
    config_loader: ConfigLoader,
    summaries: tuple,
) -> QuickStart:
    """
    Generates the 'Quick Start' section of the README file.
    """
    default_setup = QuickStart("", "", "", {}, "", "")

    try:
        language_counts = count_languages(summaries, config_loader)
        setup = (
            get_top_language_setup(language_counts, config_loader)
            or default_setup
        )

    except Exception as exc:
        _logger.debug(f"Exception: {exc}\n{traceback.format_exc()}")
        setup = default_setup

    _logger.info(f"Quickstart information: {setup}")

    return setup



"""
Custom logger implementation for the readme-ai package.
"""

import logging
import sys
import typing

LOG_LEVEL_EMOJIS = {
    "DEBUG": "⚙︎",
    "INFO": "►",
    "WARNING": "⚠️",
    "ERROR": "ⓧ",
    "CRITICAL": "‼",
}

LOG_LEVEL_COLORS = {
    "DEBUG": "\033[34m",
    "INFO": "\033[35m",
    "WARNING": "\033[33m",
    "ERROR": "\033[31m",
    "CRITICAL": "\033[31m\033[1m",
}


RESET_COLOR = "\033[0m"


class CustomFormatter(logging.Formatter):
    """
    Custom logging formatter with color and emoji support.
    """

    def format(self, record) -> str:
        """Format the log record."""
        record.emoji = LOG_LEVEL_EMOJIS.get(record.levelname, "")
        record.color = LOG_LEVEL_COLORS.get(record.levelname, "")
        super().format(record)
        return f"{record.color}{record.emoji} {record.levelname} | {record.asctime} | {record.name} | {RESET_COLOR}{record.message}"


class Logger:
    """
    Custom logger class for the readme-ai package.
    """

    _instances: typing.ClassVar[dict[str, "Logger"]] = {}

    def __new__(cls, name, level="DEBUG"):
        """Creates a new logger instance."""
        if name not in cls._instances:
            instance = super().__new__(cls)
            instance._name = name
            instance._level = level
            instance._configure_logger()
            cls._instances[name] = instance
        return cls._instances[name]

    def _configure_logger(self):
        """Configures the logger."""
        formatter = CustomFormatter(
            "%(asctime)s | %(name)s | %(levelname)s | %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        handler = logging.StreamHandler(sys.stderr)
        handler.setFormatter(formatter)
        logger = logging.getLogger(self._name)
        logger.addHandler(handler)
        logger.setLevel(self._level)

    def info(self, msg, *args, **kwargs):
        """Logs an info message."""
        logging.getLogger(self._name).info(msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        """Logs a debug message."""
        logging.getLogger(self._name).debug(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        """Logs a warning message."""
        logging.getLogger(self._name).warning(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        """Logs an error message."""
        logging.getLogger(self._name).error(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        """Logs a critical message."""
        logging.getLogger(self._name).critical(msg, *args, **kwargs)

    def log(self, level, msg, *args, **kwargs):
        """Logs a message at the specified level."""
        logging.getLogger(self._name).log(level, msg, *args, **kwargs)



"""
Factory class that selects appropriate LLM API service based on CLI input.
"""

from typing import ClassVar

from readmeai._exceptions import UnsupportedServiceError
from readmeai.config.settings import ConfigLoader, ModelOptions
from readmeai.core.models import BaseModelHandler
from readmeai.models.gemini import GeminiHandler
from readmeai.models.offline import OfflineHandler
from readmeai.models.openai import OpenAIHandler


class ModelRegistry:
    """
    Returns the appropriate LLM API handler based on CLI input.
    """

    _model_map: ClassVar[dict] = {
        # ModelOptions.ANTHROPIC.value: AnthropicHandler,
        ModelOptions.GEMINI.value: GeminiHandler,
        ModelOptions.OFFLINE.value: OfflineHandler,
        ModelOptions.OLLAMA.value: OpenAIHandler,
        ModelOptions.OPENAI.value: OpenAIHandler,
    }

    @staticmethod
    def get_backend(conf: ConfigLoader) -> BaseModelHandler:
        """
        Returns the appropriate LLM API handler based on CLI input.
        """
        backend_service = ModelRegistry._model_map.get(conf.config.llm.api)
        if backend_service is None:
            raise UnsupportedServiceError(
                f"Unsupported LLM service provided: {conf.config.llm.api}",
            )
        return backend_service(conf)



"""
CLI entrypoint for the readme-ai package.
"""

import click

from readmeai.__main__ import readme_agent
from readmeai.cli import options


@click.command()
@options.align
@options.api
@options.badge_color
@options.badge_style
@options.base_url
@options.context_window
@options.emojis
@options.header_style
@options.image
@options.model
@options.output
@options.rate_limit
@options.repository
@options.temperature
@options.toc_style
@options.top_p
@options.tree_depth
def main(
    align: str,
    api: str,
    badge_color: str,
    badge_style: str,
    base_url: str,
    context_window: int,
    emojis: bool,
    header_style: str,
    image: str,
    model: str,
    output: str,
    rate_limit: int,
    repository: str,
    temperature: float,
    toc_style: str,
    top_p: float,
    tree_depth: int,
) -> None:
    """Entry point for the readme-ai CLI application."""
    readme_agent(
        align=align,
        api=api,
        badge_color=badge_color,
        badge_style=badge_style,
        base_url=base_url,
        context_window=context_window,
        emojis=emojis,
        header_style=header_style,
        image=image,
        model=model,
        output=output,
        rate_limit=rate_limit,
        repository=repository,
        temperature=temperature,
        toc_style=toc_style,
        top_p=top_p,
        tree_depth=tree_depth,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
Main module for the README file generator agent.
"""

__package__ = "readmeai"

import asyncio
import contextlib
import tempfile
import traceback
from collections.abc import Generator

from readmeai._exceptions import ReadmeGeneratorError
from readmeai.cli.options import ImageOptions, ModelOptions
from readmeai.config.settings import ConfigLoader, GitSettings
from readmeai.core.logger import Logger
from readmeai.core.preprocess import preprocessor
from readmeai.core.utils import get_environment
from readmeai.generators.builder import MarkdownBuilder
from readmeai.models.dalle import DalleHandler
from readmeai.models.factory import ModelRegistry
from readmeai.utils.file_handler import FileHandler
from readmeai.vcs.ingestor import retrieve_repository

_logger = Logger(__name__)


@contextlib.contextmanager
def error_handler() -> Generator:
    """
    Exception handler for the README generation process.
    """
    try:
        yield
    except Exception as exc:
        _logger.error(f"Error in README generation process: {exc}")
        raise ReadmeGeneratorError(exc, traceback.format_exc()) from exc


def readme_agent(
    repository: str,
    align: str,
    api: str,
    badge_color: str,
    badge_style: str,
    base_url: str,
    context_window: int,
    emojis: bool,
    header_style: str,
    image: str,
    model: str,
    output: str,
    rate_limit: int,
    temperature: float,
    toc_style: str,
    top_p: float,
    tree_depth: int,
) -> None:
    """
    Configures and runs the README file generator agent.
    """
    with error_handler():
        api, model = get_environment(api, model)

        conf = ConfigLoader()

        conf.config.api.rate_limit = rate_limit
        conf.config.llm = conf.config.llm.model_copy(
            update={
                "api": api,
                "base_url": base_url,
                "context_window": context_window,
                "model": model,
                "temperature": temperature,
                "top_p": top_p,
            },
        )
        conf.config.md = conf.config.md.model_copy(
            update={
                "align": align,
                "badge_color": badge_color,
                "badge_style": badge_style,
                "emojis": emojis,
                "header_style": header_style,
                "image": image,
                "toc_style": toc_style,
                "tree_depth": tree_depth,
            },
        )
        conf.config.git = GitSettings(repository=repository)

        _logger.info(f"Repository settings updated: {conf.config.git}")
        _logger.info(f"LLM API settings updated: {conf.config.llm}")

        asyncio.run(readme_generator(conf, output))


async def readme_generator(conf: ConfigLoader, output_file: str) -> None:
    """
    Processes the repository and generates the README file.
    """
    with tempfile.TemporaryDirectory() as tmp_dir:
        await retrieve_repository(conf.config.git.repository, tmp_dir)
        deps, raw_files = preprocessor(conf, tmp_dir)

        _logger.info(f"Total files analyzed: {len(raw_files)}")
        _logger.info(f"Dependencies extracted: {deps}")

        async with ModelRegistry.get_backend(conf).use_api() as llm:
            responses = await llm.batch_request(deps, raw_files)
            (
                summaries,
                features,
                overview,
                slogan,
            ) = responses
            conf.config.md.features = conf.config.md.features.format(features)
            conf.config.md.overview = conf.config.md.overview.format(overview)
            conf.config.md.slogan = slogan

        if (
            conf.config.md.image == ImageOptions.LLM.value
            and conf.config.llm.api != ModelOptions.OFFLINE.value
        ):
            with DalleHandler(conf).use_api() as dalle:
                image_url = dalle._make_request()
                conf.config.md.image = dalle.download(image_url)
                if conf.config.md.image == ImageOptions.LLM.value:
                    conf.config.md.image = ImageOptions.BLUE.value

        readme_md = MarkdownBuilder(conf, deps, summaries, tmp_dir).build()

        FileHandler().write(output_file, readme_md)

        _logger.info(f"README.md file saved to: {output_file}")
        _logger.info("README generation process completed successfully!")
        _logger.info("Share it @ github.com/eli64s/readme-ai/discussions")



"""
Creates Markdown tables to store LLM text responses in the README file.
"""

from pathlib import Path

from readmeai.core.logger import Logger
from readmeai.vcs.url_builder import GitURL

_logger = Logger(__name__)


def construct_markdown_table(
    data: list[tuple[str, str]],
    repo_path: str | Path,
    full_name: str,
    max_rows: int = 100,
) -> str:
    """
    Builds a Markdown table to store LLM text responses in README file.
    """
    assert isinstance(data, list), "Data must be a list"
    assert all(
        isinstance(item, tuple) and len(item) == 2 for item in data
    ), "Each data item must be a tuple of (str, str)"
    assert isinstance(
        repo_path,
        str | Path,
    ), "repo_path must be a string or Path"
    assert isinstance(full_name, str), "full_name must be a string"
    assert (
        isinstance(max_rows, int) and max_rows > 0
    ), "max_rows must be a positive integer"

    if not data:
        _logger.warning("Empty data provided for Markdown table")
        return ""

    is_local_repo = Path(repo_path).exists()

    if not is_local_repo:
        try:
            git_url = GitURL.create(str(repo_path))
        except ValueError:
            _logger.error(f"Invalid Git repository URL: {repo_path}")
            is_local_repo = True  # Fallback to treating it as a local path

    headers = ["File", "Summary"]
    table_rows = [headers, ["---", "---"]]

    for module, summary in data[:max_rows]:
        file_name = Path(module).name
        if is_local_repo:
            file_path = Path(repo_path) / module
            md_format_file_url = f"[{file_name}]({file_path})"
        else:
            try:
                file_url = git_url.get_file_url(module)
                md_format_file_url = f"[{file_name}]({file_url})"
            except ValueError as e:
                _logger.error(f"Error generating file URL for {module}: {e}")
                md_format_file_url = file_name

        table_rows.append([md_format_file_url, summary])

    if len(data) > max_rows:
        _logger.warning(
            f"Table truncated. Showing {max_rows} out of {len(data)} rows.",
        )
        table_rows.append(["...", "..."])

    return _format_as_markdown_table(table_rows)


def _format_as_markdown_table(rows: list[list[str]]) -> str:
    """
    Formats the given rows as a Markdown table.
    """
    assert len(rows) >= 3, "Table must have at least headers and separator"
    assert all(
        len(row) == len(rows[0]) for row in rows
    ), "All rows must have the same number of columns"

    return "\n".join(f"| {' | '.join(row)} |" for row in rows)


def extract_folder_name(module: str) -> str:
    """
    Extracts the folder name from a module path.
    """
    path_parts = Path(module).parts
    return ".".join(path_parts[:-1]) if len(path_parts) > 1 else "."


def format_as_markdown_table(rows: list[list[str]]) -> str:
    """
    Formats rows of data as a Markdown table.
    """
    max_column_widths = [
        max(len(str(row[col])) for row in rows) for col in range(len(rows[0]))
    ]

    formatted_lines = [
        "| "
        + " | ".join(
            str(item).ljust(width)
            for item, width in zip(row, max_column_widths, strict=False)
        )
        + " |"
        for row in rows
    ]

    return "\n".join(formatted_lines)


def format_code_summaries(
    placeholder: str,
    code_summaries: list[tuple[str, str]],
) -> list[tuple[str, str]]:
    """
    Converts the given code summaries into a formatted list.
    """
    formatted_summaries = []

    for summary in code_summaries:
        if is_valid_tuple_summary(summary):
            module, summary_text = summary
        else:
            module, summary_text = summary, placeholder

        formatted_summaries.append((module, summary_text))

    return formatted_summaries


def generate_markdown_tables(
    table_widget: str,
    summaries: list[tuple[str, str]],
    project_name: str,
    repository_url: str,
) -> str:
    """
    Produces Markdown tables for each project sub-directory.
    """
    summaries_by_folder = group_summaries_by_folder(summaries)

    markdown_tables = []
    for folder, entries in summaries_by_folder.items():
        table_in_markdown = construct_markdown_table(
            entries,
            repository_url,
            project_name,
        )
        table_wrapper = table_widget.format(folder, table_in_markdown)
        markdown_tables.append(table_wrapper)

    return "\n".join(markdown_tables)


def group_summaries_by_folder(summaries: list[tuple[str, str]]) -> dict:
    """
    Groups code summaries by their sub-directory.
    """
    folder_map: dict[str, list[tuple[str, str]]] = {}
    for module, summary in summaries:
        folder_name = extract_folder_name(module)
        folder_map.setdefault(folder_name, []).append((module, summary))
    return folder_map


def is_valid_tuple_summary(summary: tuple[str, str]) -> bool:
    """
    Checks if a summary is a valid tuple format.
    """
    return isinstance(summary, tuple) and len(summary) == 2



"""
Abstract factory module for all project file parsers.
"""

from readmeai.core.parsers import BaseFileParser
from readmeai.parsers.cpp import (
    CMakeParser,
    ConfigureAcParser,
    MakefileAmParser,
)
from readmeai.parsers.docker import (
    DockerComposeParser,
    DockerfileParser,
)
from readmeai.parsers.go import GoModParser
from readmeai.parsers.gradle import (
    BuildGradleKtsParser,
    BuildGradleParser,
)
from readmeai.parsers.maven import MavenParser
from readmeai.parsers.npm import PackageJsonParser
from readmeai.parsers.properties import PropertiesParser
from readmeai.parsers.python import (
    RequirementsParser,
    TomlParser,
    YamlParser,
)
from readmeai.parsers.rust import CargoTomlParser
from readmeai.parsers.swift import SwiftPackageParser
from readmeai.parsers.yarn import YarnLockParser

ParserRegistryType = dict[str, BaseFileParser]


def parser_handler() -> ParserRegistryType:
    """
    Returns a dictionary of callable file parser methods.
    """
    return {
        # Python
        "Pipfile": TomlParser(),
        "pyproject.toml": TomlParser(),
        "requirements.in": RequirementsParser(),
        "requirements.txt": RequirementsParser(),
        "requirements-dev.txt": RequirementsParser(),
        "requirements-test.txt": RequirementsParser(),
        "requirements-prod.txt": RequirementsParser(),
        "dev-requirements.txt": RequirementsParser(),
        "environment.yml": YamlParser(),
        "environment.yaml": YamlParser(),
        # "setup.py": setup_py_parser,
        # "setup.cfg": setup_cfg_parser,
        # C/C++
        "cmakeLists.txt": CMakeParser(),
        "configure.ac": ConfigureAcParser(),
        "Makefile.am": MakefileAmParser(),
        # JavaScript/Node.js
        "package.json": PackageJsonParser(),
        "yarn.lock": YarnLockParser(),
        # Kotlin/Kotlin DSL
        "build.gradle": BuildGradleParser(),
        "build.gradle.kts": BuildGradleKtsParser(),
        # Go
        "go.mod": GoModParser(),
        # Java
        "pom.xml": MavenParser(),
        # Rust
        "cargo.toml": CargoTomlParser(),
        # Swift
        "Package.swift": SwiftPackageParser(),
        # Docker
        "Dockerfile": DockerfileParser(),
        "docker-compose.yaml": DockerComposeParser(),
        ".properties": PropertiesParser(),
    }



"""
Pre-processes the input repository files and extracts metadata.
"""

from collections.abc import Generator
from dataclasses import dataclass, field
from pathlib import Path

from readmeai.config.settings import ConfigLoader
from readmeai.core.logger import Logger
from readmeai.generators.builder import MarkdownBuilder
from readmeai.parsers.factory import parser_handler

_github_actions_path = ".github/workflows"


@dataclass
class FileContext:
    """
    Data class to store file contents and metadata.
    """

    file_path: Path
    file_name: str
    file_ext: str
    content: str
    language: str = field(init=False)
    dependencies: list[str] = field(default_factory=list)

    def __post_init__(self):
        """Initializes the FileContext class."""
        self.file_ext = (
            self.file_name.split(".")[-1] if "." in self.file_name else ""
        )
        self.language = self.file_ext.lower()


class RepositoryProcessor:
    """Processes the repository files and generates FileContext list"""

    def __init__(self, config_loader: ConfigLoader):
        """Initializes the RepositoryProcessor class."""
        self._logger = Logger(__name__)
        self.config_loader = config_loader
        self.ignore_list = config_loader.ignore_list.get("ignore_list")
        self.commands = config_loader.commands
        self.languages = config_loader.languages.get("language_names")
        self.parser_files = config_loader.parsers.get("parsers")

    def create_file_data(
        self,
        file_info: tuple[str, Path, str],
    ) -> FileContext:
        """Creates a FileContext instance from the file information."""
        file_name, file_path, content = file_info
        return FileContext(
            file_name=file_name,
            file_path=file_path,
            content=content,
            file_ext="",
        )

    def extract_dependencies(self, file_data: FileContext) -> list[str]:
        """Extracts the dependency file contents using the factory pattern."""
        parsers = parser_handler()
        parser = parsers.get(file_data.file_name)
        if file_data.file_name not in parsers or parser is None:
            return []
        return parser.parse(content=file_data.content)

    def generate_contents(self, repo_path: Path | str) -> list[FileContext]:
        """Generates a List of Dict of file information."""
        if isinstance(repo_path, str):
            repo_path = Path(repo_path)
        return [file_data for file_data in self.generate_file_info(repo_path)]

    def generate_file_info(
        self,
        repo_path: Path,
    ) -> Generator[FileContext, None, None]:
        """
        Generates FileContext instances for each file in the repository.
        """
        for file_path in repo_path.rglob("*"):
            if self._filter_file(file_path):
                continue
            file_data = self._process_file_path(file_path, repo_path)
            if file_data:
                yield file_data

    def get_dependencies(self, contents: list[FileContext]) -> list[str]:
        """Returns a list of dependencies."""
        try:
            dependencies = set()
            parser_files = self.parser_files

            for file_data in contents:
                dependencies.update(file_data.dependencies)
                dependencies.add(file_data.language.lower())
                dependencies.add(file_data.file_ext)

                if file_data.file_name in parser_files["parsers"]:
                    dependencies.add(file_data.file_name)
                if _github_actions_path in str(file_data.file_path):
                    dependencies.add("github actions")

            return list(dependencies)

        except Exception as exc:
            self._logger.error(f"Error getting dependencies: {exc}")
            return []

    def _filter_file(self, file_path: Path) -> bool:
        """
        Determines if a file should be ignored based on configurations.
        """
        is_file_ignored = any(
            [
                file_path.name in self.ignore_list["files"],
                file_path.suffix.lstrip(".") in self.ignore_list["extensions"],
                any(
                    dir in file_path.parts
                    for dir in self.ignore_list["directories"]
                ),
            ],
        )
        if is_file_ignored and str(file_path.name) in self.parser_files:
            return False
        return not file_path.is_file() or is_file_ignored

    def _language_mapper(
        self,
        contents: list[FileContext],
    ) -> list[FileContext]:
        """Maps file extensions to their programming languages."""
        for content in contents:
            content.language = self.languages.get(content.file_ext, "").lower()
        return contents

    def _process_file_path(
        self,
        file_path: Path,
        repo_path: Path,
    ) -> FileContext:
        """
        Processes an individual file path and returns FileContext.
        """
        relative_path = file_path.relative_to(repo_path)
        if _github_actions_path in str(relative_path):
            return FileContext(
                file_path=relative_path,
                file_name="github actions",
                file_ext="",
                content="",
            )
        try:
            if file_path.is_dir():
                return FileContext(
                    file_path=relative_path,
                    file_name=file_path.name,
                    file_ext="",
                    content="",
                )
            with file_path.open(encoding="utf-8") as file:
                content = file.read()

            file_data = FileContext(
                file_path=relative_path,
                file_name=file_path.name,
                file_ext=file_path.suffix,
                content=content,
            )
            file_data.dependencies = self.extract_dependencies(file_data)
            return file_data

        except (OSError, UnicodeDecodeError) as exc:
            self._logger.warning(f"Error reading file {file_path}: {exc}")
            return FileContext(
                file_path=relative_path,
                file_name=file_path.name,
                file_ext="",
                content="",
            )


def preprocessor(conf: ConfigLoader, tmp_dir: str) -> tuple:
    """Processes the repository files and returns the context."""
    repo_processor = RepositoryProcessor(conf)
    repo_context = repo_processor.generate_contents(tmp_dir)
    repo_context = repo_processor._language_mapper(repo_context)
    deps = repo_processor.get_dependencies(repo_context)
    raw_files = [
        (str(context.file_path), context.content) for context in repo_context
    ]
    conf.config.md.tree = MarkdownBuilder(
        conf,
        deps,
        raw_files,
        tmp_dir,
    ).md_tree
    return deps, raw_files



"""
Custom exceptions classes for the readme-ai package.
"""

from __future__ import annotations


class ReadmeAIError(Exception):
    """
    Base class for exceptions in this module.
    """

    ...


class ReadmeGeneratorError(Exception):
    """
    Raised when an error occurs during README generation.
    """

    def __init__(self, exc, traceback):
        self.exc = exc
        self.traceback = traceback
        super().__init__(f"README generation error occurred: {exc}")


# ----------------- CLI ----------------------------------


class CLIError(ReadmeAIError):
    """Exceptions related to the CLI."""

    def __init__(self, message, *args):
        super().__init__(f"Invalid option provided to CLI: {message}", *args)


# ----------------- File System ----------------------------------


class FileSystemError(ReadmeAIError):
    """
    Exceptions related to file system operations.
    """

    def __init__(self, message, path, *args):
        self.file_path = path
        super().__init__(f"{message}: {path}", *args)


class FileReadError(FileSystemError):
    """
    Raised when a file cannot be read.
    """

    ...


class FileWriteError(FileSystemError):
    """
    Raised when a file cannot be written to.
    """

    ...


# ----------------- LLM API ----------------------------------


class UnsupportedServiceError(ReadmeAIError):
    """
    Raised when an unsupported LLM service is provided.
    """

    def __init__(self, message, *args):
        super().__init__(message, *args)



from abc import ABC, abstractmethod
from typing import Any


class BaseTemplate(ABC):
    """
    Base class for all templates.
    """

    @abstractmethod
    def render(self, data: dict[str, Any]) -> str:
        """
        Render the template with the given data.

        Args:
            data (Dict[str, Any]): The data to use for rendering the template.

        Returns:
            str: The rendered template as a string.
        """
        ...

    @staticmethod
    def sanitize_input(input_str: str) -> str:
        """
        Sanitize input string to prevent XSS attacks.

        Args:
            input_str (str): The input string to sanitize.

        Returns:
            str: The sanitized string.
        """
        return input_str.replace("<", "&lt;").replace(">", "&gt;")



from enum import Enum
from pathlib import Path

from pydantic import HttpUrl

from .errors import GitURLError


class GitHost(Enum):
    """
    Enum for supported Git repository hosting services.
    """

    GITHUB = (
        "github.com",
        "https://api.github.com/repos/",
        "https://github.com/{full_name}/blob/main/{file_path}",
    )
    GITLAB = (
        "gitlab.com",
        "https://gitlab.com/api/v4/projects/",
        "https://gitlab.com/{full_name}/-/blob/main/{file_path}",
    )
    BITBUCKET = (
        "bitbucket.org",
        "https://api.bitbucket.org/2.0/repositories/",
        "https://bitbucket.org/{full_name}/src/master/{file_path}",
    )
    LOCAL = ("local", "", "{file_path}")

    def __init__(self, name: str, api_url: str, file_url_template: str):
        """
        Initialize git host domain, REST API URL, and file URL template.
        """
        self.domain = name
        self.api_url = api_url
        self.file_url_template = file_url_template


def parse_git_url(url: str | Path) -> tuple[str, str, str, str]:
    """
    Parse Git repository URL and return host, full name, and project name.
    """
    if isinstance(url, Path) or (isinstance(url, str) and Path(url).is_dir()):
        host_domain = host = GitHost.LOCAL.name
        name = Path(url).name
        full_name = f"{Path(url).parent.name}/{name}"
    else:
        try:
            parsed_url = HttpUrl(url)
            if parsed_url.scheme not in ["http", "https"]:
                raise GitURLError(
                    url,
                    f"Uknown scheme provided: {parsed_url.scheme}",
                )
        except ValueError as e:
            raise GitURLError(url) from e

        assert (
            parsed_url.host and parsed_url.path
        ), f"Invalid Git repository URL: {parsed_url}"
        path_parts = parsed_url.path.strip("/").split("/")
        host_domain = parsed_url.host
        host = parsed_url.host.split(".")[0].lower()
        name = path_parts[-1]
        full_name = "/".join(path_parts[:2])

    return host_domain, host, name, full_name


def get_file_url(host: GitHost, full_name: str, file_path: str) -> str:
    """
    Return the URL of the file in the remote repository.
    """
    return host.file_url_template.format(
        full_name=full_name,
        file_path=file_path,
    )



import functools

from pydantic import (
    BaseModel,
    Field,
    HttpUrl,
    field_validator,
    model_validator,
)

from .providers import GitHost, get_file_url, parse_git_url


class GitURL(BaseModel):
    """
    Git repository URL model with validation and parsing methods.
    """

    url: HttpUrl
    host: GitHost | None = Field(default=None)
    host_domain: str = Field(default="")
    name: str = Field(default="")
    full_name: str = Field(default="")

    model_config = {
        "frozen": True,
        "use_enum_values": True,
        "extra": "forbid",
        "arbitrary_types_allowed": True,
    }

    @field_validator("url")
    @classmethod
    def validate_url(cls, v: HttpUrl) -> HttpUrl:
        """
        Validates the Git repository URL.
        """
        try:
            parse_git_url(str(v))
        except ValueError as e:
            raise ValueError(f"Invalid Git repository URL: {v}") from e
        return v

    @model_validator(mode="after")
    def set_attributes(self) -> "GitURL":
        """
        Sets the Git URL attributes based on the URL.
        """
        try:
            host_domain, host, name, full_name = parse_git_url(str(self.url))
            object.__setattr__(self, "host_domain", host_domain)
            object.__setattr__(self, "name", name)
            object.__setattr__(self, "full_name", full_name)
            for git_host in GitHost:
                if git_host.name.lower() == host:
                    object.__setattr__(self, "host", git_host)
                    break
        except ValueError as e:
            raise ValueError(f"Failed to parse Git URL: {self.url}") from e
        return self

    @classmethod
    @functools.lru_cache(maxsize=100)
    def create(cls, url: HttpUrl) -> "GitURL":
        """
        Create a GitURL object from a string URL.
        """
        return cls(url=url)

    def get_api_url(self) -> str:
        """
        Return the REST API endpoint URL for a git repository.
        """
        # Ensure self.host is not None and self.host is not GitHost.LOCAL
        if self.host is not None and self.host != GitHost.LOCAL:
            # Ensure self.full_name is not None or empty
            if self.full_name:
                return f"{self.host.api_url}/{self.full_name}"
            else:
                raise ValueError("Repository full name is required.")
        else:
            raise ValueError(
                f"Unsupported Git host or local repository: {self.url}",
            )

    def get_file_url(self, file_path: str) -> str:
        """
        Return the URL of the file in the remote repository.
        """
        if self.host:
            return get_file_url(self.host, self.full_name, file_path)
        raise ValueError(f"Unsupported Git host: {self.url}")



"""
Utility methods to read package resource files.
"""

from importlib import resources
from pathlib import Path

from readmeai._exceptions import FileReadError


def get_resource_path(
    file_path: str,
    package: str = "readmeai.config",
    submodule: str = "settings",
) -> str:
    """Retrieves the path to a resource file within the package.

    This function attempts to first use `importlib.resources` for preferred
    access to resources within the package. It falls back to `pkg_resources`
    for compatibility with older environments.

    Parameters
    ----------
    file_path
        The path to the resource file relative to the package's submodule.
    package, optional
        The package name containing the resource file.
        - default: "readmeai.config"
    submodule, optional
        The submodule within the package where the resource is located.
        - default: "settings"

    Returns
    -------
        The absolute path to the resource file.

    Raises
    ------
    FileReadError
        FileReadError: If the resource file cannot be found or accessed.
    """
    resource_path = None
    try:
        resource_path = resources.files(package).joinpath(submodule, file_path)

    except TypeError:  # pragma: no cover
        try:
            import pkg_resources

            submodule = submodule.replace(".", "/")
            resource_path = Path(
                pkg_resources.resource_filename(
                    "readmeai",
                    f"{submodule}/{file_path}",
                ),
            ).resolve()

        except Exception as exc:  # pragma: no cover
            raise FileReadError(
                "Error loading resource file via pkg_resources",
                str(resource_path),
            ) from exc

    if resource_path is None:
        raise FileReadError("Resource file not found", str(resource_path))

    return str(resource_path)



"""
Utility methods to configure the LLM API environments.
"""

import os
from enum import Enum

from readmeai.config.settings import ModelOptions as llms
from readmeai.core.logger import Logger

_logger = Logger(__name__)


class SecretKey(str, Enum):
    """
    Enum class to store the environment variable keys for the LLM API services.
    """

    # ANTHROPIC_API_KEY = "ANTHROPIC_API_KEY"
    GOOGLE_API_KEY = "GOOGLE_API_KEY"
    OLLAMA_HOST = "OLLAMA_HOST"
    OPENAI_API_KEY = "OPENAI_API_KEY"


def _set_offline(message: str) -> tuple:
    """Set the LLM service to offline mode."""
    _logger.warning(
        f"{message}\n\t\n... readme-ai engine switched to offline mode ...\n",
    )
    return llms.OFFLINE.name, llms.OFFLINE.name


def get_environment(llm_api: str = "", llm_model: str = "") -> tuple:
    """Set LLM environment variables based on the specified LLM service."""
    default_models = {
        # llms.ANTHROPIC.name: "claude-3-5-sonnet",
        llms.GEMINI.name: "gemini-1.5-flash",
        llms.OLLAMA.name: "mistral",
        llms.OPENAI.name: "gpt-3.5-turbo",
    }

    env_keys = {
        # llms.ANTHROPIC.name: SecretKey.ANTHROPIC_API_KEY.value,
        llms.GEMINI.name: SecretKey.GOOGLE_API_KEY.value,
        llms.OLLAMA.name: SecretKey.OLLAMA_HOST.value,
        llms.OPENAI.name: SecretKey.OPENAI_API_KEY.value,
    }

    if llm_api and llm_api not in env_keys:
        if llm_api == llms.OFFLINE.name:
            return _set_offline("\n\n\t\t\t\tOffline mode enabled by user")
        _logger.warning("Invalid LLM service provided to CLI.")
        return _set_offline(
            "\n\n\t...No LLM API settings found in environment...",
        )

    # If OPENAI_API_KEY does not exist in env when --api OPENAI is set
    if (
        llm_api == llms.OPENAI.name
        and SecretKey.OPENAI_API_KEY.value not in os.environ
    ):
        return _set_offline(
            "OPENAI_API_KEY not found in environment. Switching to offline mode.",
        )

    # If GOOGLE_API_KEY does not exist in env when --api gemini is set
    if (
        llm_api == llms.GEMINI.name
        and SecretKey.GOOGLE_API_KEY.value not in os.environ
    ):
        return _set_offline(
            "GOOGLE_API_KEY not found in environment. Switching to offline mode.",
        )

    # If no specific API is provided or the provided API is valid
    for api_name, env_key in env_keys.items():
        if llm_api == api_name or (not llm_api and env_key in os.environ):
            model = llm_model if llm_model else default_models[api_name]
            _logger.info(f"{api_name} settings FOUND in environment!")
            return api_name, model

    # If no environment variables are found or OFFLINE is explicitly set
    if llm_api == llms.OFFLINE.name:
        return _set_offline("Offline mode enabled by user via CLI.")
    return _set_offline(
        "\n\n... No LLM API settings found in environment ...",
    )



"""
Handler for generating images using OpenAI's DALL-E model.
"""

import os
from collections.abc import Generator
from contextlib import contextmanager

from openai import Client, OpenAIError
from requests import get

from readmeai.cli.options import ImageOptions
from readmeai.config.settings import ConfigLoader
from readmeai.core.logger import Logger


class DalleHandler:
    """
    Generates and downloads images using OpenAI's DALL-E model.
    """

    def __init__(self, conf: ConfigLoader) -> None:
        """Initialize the ImageGenerator class."""
        self.conf = conf
        self.filename = f"{conf.config.git.name}.png"
        self._logger = Logger(__name__)
        self._model_settings()

    def _model_settings(self) -> None:
        """Initializes the DALL-E settings."""
        self.client = Client(api_key=os.getenv("OPENAI_API_KEY"))
        self.model = "dall-e-3"
        self.size = "1792x1024"
        self.quality = "standard"
        self.n = 1

    @contextmanager
    def use_api(self) -> Generator:
        """Yields the DALL-E handler."""
        try:
            yield self
        finally:
            self._logger.debug(f"Closed {self.model.upper()} API session.")

    def _build_payload(self) -> str:
        """Formats the prompt string using configuration data."""
        return {
            "prompt": self.conf.prompts["prompts"]["logo"].format(
                project_name=self.conf.config.git.name,
                project_slogan=self.conf.config.md.slogan,
                project_overview=self.conf.config.md.overview,
            ),
            "model": self.model,
            "size": self.size,
            "quality": self.quality,
            "n": self.n,
        }

    def _make_request(self) -> str:
        """Generates an image and returns its URL."""
        try:
            payload = self._build_payload()
            response = self.client.images.generate(**payload)
            if response is not None:
                return response.data[0].url
            else:
                self._logger.error(
                    f"Failed to generate {self.model.upper()} image: {response}",
                )
                return ImageOptions.BLUE.value

        except (Exception, OpenAIError) as exc:
            self._logger.error(
                f"{self.model.upper()} image generation error: {exc}",
            )
            return ImageOptions.BLUE.value

    def download(self, image_url) -> str:
        """Downloads an image from the given URL."""
        try:
            response = get(image_url)
            status_code = response.status_code

            if status_code == 200:
                with open(self.filename, "wb") as f:
                    f.write(response.content)
                self._logger.info(f"Image downloaded at: {image_url}")
                return self.filename
            else:
                self._logger.error(f"Failed to download image: {status_code}")

        except Exception as exc:
            self._logger.error(f"Failed to download image: {exc}")

        return ImageOptions.BLUE.value



import enum
from typing import Any, ClassVar

from .base_template import BaseTemplate


class HeaderStyle(enum.StrEnum):
    """
    Enum containing header styles for the README file.
    """

    CLASSIC = enum.auto()
    COMPACT = enum.auto()
    MODERN = enum.auto()


class HeaderTemplate(BaseTemplate):
    """
    Class variable for rendering the README.md header style.
    """

    HEADER_TEMPLATES: ClassVar[dict] = {
        HeaderStyle.CLASSIC: """\
<p align="{align}">
  <img src="{image}" width="{image_width}" alt="{repo_name}-logo">
</p>
<p align="{align}">
    <h1 align="{align}">{repo_name}</h1>
</p>
<p align="{align}">
    <em>{slogan}</em>
</p>
<p align="{align}">\n\t{shields_icons}</p>
<p align="{align}">\n\t{badge_icons}</p>

<br>
""",
        HeaderStyle.COMPACT: """\
[<img src="{image}" align="left" width="{image_width}" padding="20">]()

## &nbsp;&nbsp; {repo_name}

&nbsp;&nbsp;&nbsp;&nbsp; *{slogan}*

<p align="left">&nbsp;&nbsp;\n\t{shields_icons}</p>

<br>
""",
        HeaderStyle.MODERN.value: """\
[<img src="{image}" align="right" width="25%" padding-right="350">]()

# `{repo_name}`

#### {slogan}

<p align="left">\n\t{shields_icons}</p>
<p align="left">\n\t{badge_icons}</p>

<br>
""",
    }

    def __init__(self, style: str = HeaderStyle.CLASSIC) -> None:
        """
        Initialize the header template with the given style.
        """
        self.style = style

    def render(self, data: dict[str, Any]) -> str:
        """
        Render the header based on the current style and given data.
        """
        template = self.HEADER_TEMPLATES.get(
            self.style,
            self.HEADER_TEMPLATES[HeaderStyle.CLASSIC],
        )
        return template.format(**data)

    @staticmethod
    def get_header_template(template: str) -> str:
        """
        Get the header template for the given style.
        """
        try:
            return HeaderTemplate.HEADER_TEMPLATES[HeaderStyle(template)]
        except ValueError:
            return HeaderTemplate.HEADER_TEMPLATES[HeaderStyle.CLASSIC]



"""
Google Cloud's Gemini API handler implementation.
"""

import os

import aiohttp
import google.generativeai as genai
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from readmeai.config.settings import ConfigLoader
from readmeai.core.logger import Logger
from readmeai.core.models import BaseModelHandler
from readmeai.models.tokens import token_handler
from readmeai.utils.text_cleaner import clean_response


class GeminiHandler(BaseModelHandler):
    """Google Gemini API LLM implementation."""

    def __init__(self, config_loader: ConfigLoader) -> None:
        """Initializes the Gemini API handler."""
        super().__init__(config_loader)
        self._logger = Logger(__name__)
        self._model_settings()

    def _model_settings(self):
        """Initializes the Gemini API LLM settings."""
        api_key = os.getenv("GOOGLE_API_KEY")
        genai.configure(api_key=api_key)
        self.model = genai.GenerativeModel(self.config.llm.model)
        self.temperature = self.config.llm.temperature
        self.tokens = self.config.llm.tokens
        self.top_p = self.config.llm.top_p

    async def _build_payload(self, prompt: str, tokens: int) -> dict:
        """Build payload for POST request to the Gemini API."""
        return genai.types.GenerationConfig(
            max_output_tokens=self.tokens,
            temperature=self.temperature,
            top_p=self.top_p,
        )

    @retry(
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=4, max=10),
        retry=retry_if_exception_type(
            (
                aiohttp.ClientError,
                aiohttp.ClientResponseError,
                aiohttp.ClientConnectorError,
            ),
        ),
    )
    async def _make_request(
        self,
        index: str | None,
        prompt: str | None,
        tokens: int | None,
        raw_files: list[tuple[str, str]] | None,
    ) -> list[tuple[str, str]]:
        """
        Processes Gemini API responses and returns generated text.
        """
        try:
            prompt = await token_handler(self.config, index, prompt, tokens)

            parameters = await self._build_payload(prompt, tokens)

            async with self.rate_limit_semaphore:
                response = await self.model.generate_content_async(
                    prompt,
                    generation_config=parameters,
                )
                response_text = response.text

                self._logger.info(f"Response for '{index}':\n{response_text}")

                return index, clean_response(index, response_text)

        except (
            aiohttp.ClientError,
            aiohttp.ClientResponseError,
            aiohttp.ClientConnectorError,
        ):
            self._logger.error(
                f"Error processing request for prompt: {index}",
                exc_info=True,
            )
            return index, self.config.md.placeholder
